from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Callable

from opavm import catalog, config, download, github, platform
from opavm.errors import OpavmError, VersionNotInstalledError

StatusCallback = Callable[[str], None]
DownloadProgressCallback = Callable[[int | None, int], None]


def _tool_versions_dir(tool: str) -> Path:
    return config.tool_versions_dir(tool)


def _asset_candidates(tool: str, os_name: str, arch: str) -> list[str]:
    if tool == "opa":
        return platform.asset_name_candidates(os_name, arch, binary_base="opa")

    if tool == "regal":
        os_token = {"darwin": "Darwin", "linux": "Linux", "windows": "Windows"}[os_name]
        arch_token = {"amd64": "x86_64", "arm64": "arm64"}[arch]
        suffix = ".exe" if os_name == "windows" else ""
        return [f"regal_{os_token}_{arch_token}{suffix}"]

    raise OpavmError("Unknown tool.", f"Supported tools: {', '.join(sorted(catalog.SUPPORTED_TOOLS))}")


def installed_versions(tool: str = "opa") -> list[str]:
    spec = catalog.get_tool(tool)
    root = _tool_versions_dir(spec.name)
    if not root.exists():
        return []
    versions: list[str] = []
    for entry in root.iterdir():
        if not entry.is_dir():
            continue
        if (entry / spec.binary_base).exists() or (entry / f"{spec.binary_base}.exe").exists():
            versions.append(entry.name)
    return sorted(versions)


def _platform_binary_path(version: str, os_name: str, tool: str = "opa") -> Path:
    spec = catalog.get_tool(tool)
    return _tool_versions_dir(spec.name) / version / platform.binary_filename(os_name, spec.binary_base)


def binary_path(version: str, tool: str = "opa") -> Path:
    spec = catalog.get_tool(tool)
    os_name, _ = platform.normalized_os_arch()
    preferred = _platform_binary_path(version, os_name, spec.name)
    if preferred.exists():
        return preferred
    alternate_name = (
        f"{spec.binary_base}.exe" if preferred.name == spec.binary_base else spec.binary_base
    )
    alternate = preferred.parent / alternate_name
    if alternate.exists():
        return alternate
    return preferred


def is_installed(version: str, tool: str = "opa") -> bool:
    return binary_path(version, tool=tool).exists()


def verify_binary(binary: Path) -> None:
    try:
        subprocess.run([str(binary), "version"], check=True, capture_output=True, text=True)
    except (OSError, subprocess.CalledProcessError) as exc:
        raise OpavmError("Installed binary failed verification.", "Run install again.") from exc


def install(
    version: str,
    tool: str = "opa",
    on_status: StatusCallback | None = None,
    on_download: DownloadProgressCallback | None = None,
) -> str:
    spec = catalog.get_tool(tool)
    config.ensure_layout()
    if on_status is not None:
        on_status("resolving")

    os_name, arch = platform.normalized_os_arch()
    repo = spec.default_repo
    release = github.fetch_release(version, repo=repo)
    resolved_version = release.version

    if is_installed(resolved_version, tool=spec.name):
        if on_status is not None:
            on_status("already_installed")
        return resolved_version

    expected_assets = _asset_candidates(spec.name, os_name, arch)
    raw_assets = getattr(release, "assets", [])
    try:
        assets = list(raw_assets)
    except TypeError:
        assets = []
    selected_asset_name = ""
    for candidate in expected_assets:
        for asset in assets:
            if asset.name == candidate:
                selected_asset_name = candidate
                break
        if selected_asset_name:
            break
    if not selected_asset_name and expected_assets:
        selected_asset_name = expected_assets[0]
    asset_url = github.pick_asset_url(release, expected_assets)

    target = _platform_binary_path(resolved_version, os_name, spec.name)
    if on_status is not None:
        on_status("downloading")
    download.download_binary(asset_url, target, on_progress=on_download)
    checksum_url = github.checksum_asset_url(release, selected_asset_name)
    if checksum_url:
        if on_status is not None:
            on_status("verifying_checksum")
        expected_checksum = download.parse_checksum_text(download.fetch_text(checksum_url))
        actual_checksum = download.sha256_file(target)
        if expected_checksum != actual_checksum:
            raise OpavmError(
                "Checksum verification failed.",
                f"Downloaded file hash mismatch for {selected_asset_name}.",
            )
    if on_status is not None:
        on_status("verifying")
    verify_binary(target)
    if on_status is not None:
        on_status("done")

    return resolved_version


def uninstall(version: str, tool: str = "opa") -> None:
    spec = catalog.get_tool(tool)
    if not is_installed(version, tool=spec.name):
        install_hint = f"opavm install {version}" if spec.name == "opa" else f"opavm install {spec.name} {version}"
        raise VersionNotInstalledError(
            "Version not installed.", f"Run: {install_hint}"
        )
    download.remove_tree(_tool_versions_dir(spec.name) / version)



from __future__ import annotations

import json
import os
import tempfile
from pathlib import Path
from typing import Any

from opavm.errors import OpavmError

DEFAULT_BASE_DIR = Path.home() / ".opavm"


def base_dir() -> Path:
    return Path(os.environ.get("OPAVM_HOME", DEFAULT_BASE_DIR)).expanduser()


def versions_dir() -> Path:
    return base_dir() / "versions"


def tool_versions_dir(tool: str) -> Path:
    if tool == "opa":
        return versions_dir()
    return base_dir() / "tools" / tool / "versions"


def shims_dir() -> Path:
    return base_dir() / "shims"


def state_path() -> Path:
    return base_dir() / "state.json"


def ensure_layout() -> None:
    versions_dir().mkdir(parents=True, exist_ok=True)
    shims_dir().mkdir(parents=True, exist_ok=True)


def _default_state() -> dict[str, Any]:
    return {"global_default": None, "global_defaults": {}}


def load_state() -> dict[str, Any]:
    path = state_path()
    if not path.exists():
        return _default_state()
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except json.JSONDecodeError as exc:
        raise OpavmError("Corrupt state file.", "Delete ~/.opavm/state.json and retry.") from exc
    if not isinstance(raw, dict):
        return _default_state()

    state = _default_state()
    state["global_default"] = raw.get("global_default")

    global_defaults = raw.get("global_defaults")
    if isinstance(global_defaults, dict):
        state["global_defaults"] = {
            str(tool): str(version)
            for tool, version in global_defaults.items()
            if isinstance(tool, str) and isinstance(version, str)
        }
    else:
        state["global_defaults"] = {}
    return state


def save_state(data: dict[str, Any]) -> None:
    ensure_layout()
    path = state_path()
    fd, tmp_name = tempfile.mkstemp(dir=str(path.parent), prefix="state.", suffix=".tmp")
    try:
        with os.fdopen(fd, "w", encoding="utf-8") as f:
            json.dump(data, f)
            f.write("\n")
            f.flush()
            os.fsync(f.fileno())
        os.replace(tmp_name, path)
    finally:
        if os.path.exists(tmp_name):
            os.remove(tmp_name)


def get_global_default(tool: str = "opa") -> str | None:
    state = load_state()
    global_defaults = state.get("global_defaults")
    if isinstance(global_defaults, dict):
        value = global_defaults.get(tool)
        if isinstance(value, str) and value.strip():
            return value

    if tool == "opa":
        legacy = state.get("global_default")
        if isinstance(legacy, str) and legacy.strip():
            return legacy
    return None


def set_global_default(tool: str, version: str) -> None:
    state = load_state()
    global_defaults = state.get("global_defaults")
    if not isinstance(global_defaults, dict):
        global_defaults = {}
    global_defaults[tool] = version
    state["global_defaults"] = global_defaults
    if tool == "opa":
        state["global_default"] = version
    save_state(state)



from __future__ import annotations

from unittest import mock

import pytest

from opavm.errors import UnsupportedPlatformError
from opavm.platform import asset_name, asset_name_candidates, binary_filename, normalized_os_arch


@pytest.mark.parametrize(
    ("sys_name", "machine", "expected"),
    [
        ("Darwin", "x86_64", ("darwin", "amd64")),
        ("Darwin", "arm64", ("darwin", "arm64")),
        ("Linux", "amd64", ("linux", "amd64")),
        ("Linux", "aarch64", ("linux", "arm64")),
        ("Windows", "x86_64", ("windows", "amd64")),
    ],
)
def test_normalized_os_arch_supported(sys_name: str, machine: str, expected: tuple[str, str]) -> None:
    with mock.patch("platform.system", return_value=sys_name), mock.patch(
        "platform.machine", return_value=machine
    ):
        assert normalized_os_arch() == expected


def test_normalized_os_arch_rejects_windows_arm64() -> None:
    with mock.patch("platform.system", return_value="Windows"), mock.patch(
        "platform.machine", return_value="arm64"
    ):
        with pytest.raises(UnsupportedPlatformError):
            normalized_os_arch()


def test_asset_name() -> None:
    assert asset_name("0.62.1", "linux", "amd64") == "opa_linux_amd64"
    assert asset_name("0.62.1", "windows", "amd64") == "opa_windows_amd64.exe"
    assert asset_name("0.38.1", "linux", "amd64", binary_base="regal") == "regal_linux_amd64"


def test_asset_name_candidates_arm64_static_fallback() -> None:
    assert asset_name_candidates("linux", "arm64") == ["opa_linux_arm64", "opa_linux_arm64_static"]


def test_binary_filename() -> None:
    assert binary_filename("linux") == "opa"
    assert binary_filename("windows") == "opa.exe"
    assert binary_filename("windows", binary_base="regal") == "regal.exe"



from __future__ import annotations

import os
import shutil
import tempfile
import hashlib
from pathlib import Path
from typing import Callable

import httpx

from opavm.errors import DownloadError


def download_binary(
    url: str,
    destination: Path,
    timeout: float = 120.0,
    on_progress: Callable[[int | None, int], None] | None = None,
) -> None:
    destination.parent.mkdir(parents=True, exist_ok=True)

    fd, temp_name = tempfile.mkstemp(dir=str(destination.parent), prefix="opa.", suffix=".tmp")
    os.close(fd)
    temp_path = Path(temp_name)

    try:
        with httpx.stream("GET", url, timeout=timeout, follow_redirects=True) as response:
            response.raise_for_status()
            total_header = response.headers.get("Content-Length")
            total_bytes = int(total_header) if total_header and total_header.isdigit() else None
            downloaded_bytes = 0
            if on_progress is not None:
                on_progress(total_bytes, downloaded_bytes)
            with temp_path.open("wb") as fh:
                for chunk in response.iter_bytes():
                    if chunk:
                        fh.write(chunk)
                        downloaded_bytes += len(chunk)
                        if on_progress is not None:
                            on_progress(total_bytes, downloaded_bytes)
                fh.flush()
                os.fsync(fh.fileno())

        temp_path.chmod(temp_path.stat().st_mode | 0o111)
        os.replace(temp_path, destination)
    except httpx.HTTPError as exc:
        raise DownloadError("Download failed.", f"Could not fetch: {url}") from exc
    except OSError as exc:
        raise DownloadError("Install failed.", "Could not place downloaded binary.") from exc
    finally:
        if temp_path.exists():
            temp_path.unlink()


def remove_tree(path: Path) -> None:
    if path.exists():
        shutil.rmtree(path)


def sha256_file(path: Path) -> str:
    digest = hashlib.sha256()
    with path.open("rb") as f:
        while True:
            chunk = f.read(1024 * 1024)
            if not chunk:
                break
            digest.update(chunk)
    return digest.hexdigest()


def fetch_text(url: str, timeout: float = 30.0) -> str:
    try:
        response = httpx.get(url, timeout=timeout, follow_redirects=True)
        response.raise_for_status()
        return response.text
    except httpx.HTTPError as exc:
        raise DownloadError("Checksum fetch failed.", f"Could not fetch: {url}") from exc


def parse_checksum_text(text: str) -> str:
    for line in text.splitlines():
        line = line.strip()
        if not line:
            continue
        token = line.split()[0]
        if len(token) == 64 and all(ch in "0123456789abcdefABCDEF" for ch in token):
            return token.lower()
    raise DownloadError("Invalid checksum file.", "No SHA256 value found.")



from __future__ import annotations

import json
from pathlib import Path

import pytest

from opavm import config


def test_load_state_defaults_when_missing(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    assert config.load_state() == {"global_default": None, "global_defaults": {}}


def test_save_state_is_atomic(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    config.save_state({"global_default": "0.62.1"})

    path = config.state_path()
    assert path.exists()
    data = json.loads(path.read_text(encoding="utf-8"))
    assert data["global_default"] == "0.62.1"

    leftovers = list(path.parent.glob("state.*.tmp"))
    assert leftovers == []


def test_set_get_global_default_by_tool(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    config.set_global_default("opa", "0.62.1")
    config.set_global_default("regal", "0.38.1")

    assert config.get_global_default("opa") == "0.62.1"
    assert config.get_global_default("regal") == "0.38.1"


def test_get_global_default_opa_falls_back_to_legacy_key(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    config.save_state({"global_default": "0.60.0"})

    assert config.get_global_default("opa") == "0.60.0"



from __future__ import annotations

from pathlib import Path
from unittest import mock

import pytest

from opavm import download, github, installer
from opavm.errors import OpavmError


def test_parse_checksum_text() -> None:
    text = "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824  opa_linux_amd64\n"
    parsed = download.parse_checksum_text(text)
    assert parsed == "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"


def test_sha256_file(tmp_path: Path) -> None:
    file_path = tmp_path / "bin"
    file_path.write_bytes(b"hello")
    assert download.sha256_file(file_path) == "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"


def test_installer_checksum_mismatch_fails(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    release = github.ReleaseInfo(
        version="0.62.1",
        tag="v0.62.1",
        assets=[
            github.ReleaseAsset(name="opa_linux_amd64", url="https://example.test/opa"),
            github.ReleaseAsset(name="opa_linux_amd64.sha256", url="https://example.test/opa.sha256"),
        ],
    )
    with mock.patch("opavm.installer.platform.normalized_os_arch", return_value=("linux", "amd64")), mock.patch(
        "opavm.installer.github.fetch_release", return_value=release
    ) as fetch_release_mock, mock.patch(
        "opavm.installer.github.pick_asset_url", return_value="https://example.test/opa"
    ), mock.patch("opavm.installer.download.fetch_text", return_value="0" * 64), mock.patch(
        "opavm.installer.verify_binary"
    ):
        def fake_download(_url: str, destination: Path, on_progress=None) -> None:
            destination.parent.mkdir(parents=True, exist_ok=True)
            destination.write_bytes(b"different-content")
            if on_progress is not None:
                on_progress(17, 17)

        with mock.patch("opavm.installer.download.download_binary", side_effect=fake_download):
            with pytest.raises(OpavmError, match="Checksum verification failed"):
                installer.install("0.62.1")
    assert fetch_release_mock.call_args.kwargs["repo"] == "open-policy-agent/opa"


def test_installer_checksum_match_passes(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    release = github.ReleaseInfo(
        version="0.62.1",
        tag="v0.62.1",
        assets=[
            github.ReleaseAsset(name="opa_linux_amd64", url="https://example.test/opa"),
            github.ReleaseAsset(name="opa_linux_amd64.sha256", url="https://example.test/opa.sha256"),
        ],
    )
    with mock.patch("opavm.installer.platform.normalized_os_arch", return_value=("linux", "amd64")), mock.patch(
        "opavm.installer.github.fetch_release", return_value=release
    ) as fetch_release_mock, mock.patch(
        "opavm.installer.github.pick_asset_url", return_value="https://example.test/opa"
    ), mock.patch("opavm.installer.verify_binary") as verify_mock:
        def fake_download(_url: str, destination: Path, on_progress=None) -> None:
            destination.parent.mkdir(parents=True, exist_ok=True)
            destination.write_bytes(b"trusted")
            if on_progress is not None:
                on_progress(7, 7)

        expected_checksum = "a9a089195c68d2adeee23beaa2c3a93b1d4cdf09046e7a9e520b3b166dff3e6a"
        with mock.patch("opavm.installer.download.download_binary", side_effect=fake_download), mock.patch(
            "opavm.installer.download.fetch_text", return_value=f"{expected_checksum}  opa_linux_amd64\n"
        ):
            installed = installer.install("0.62.1")

    assert installed == "0.62.1"
    assert fetch_release_mock.call_args.kwargs["repo"] == "open-policy-agent/opa"
    verify_mock.assert_called_once()



from __future__ import annotations

import subprocess
from pathlib import Path

import typer
from rich import box
from rich.console import Console
from rich.filesize import decimal
from rich.progress import BarColumn, Progress, SpinnerColumn, TaskProgressColumn, TextColumn
from rich.table import Table

from opavm import catalog, config, github, installer, resolver, runner, shim
from opavm.errors import OpavmError, VersionNotInstalledError

app = typer.Typer(
    no_args_is_help=True,
    help=(
        "A dead-simple OPA version manager.\n\n"
        "[bold magenta]Tool Selection[/]\n"
        "- [cyan]opa[/cyan] (default)\n"
        "- [cyan]regal[/cyan]\n\n"
        "Use [cyan]--tool[/cyan] on tool-aware commands: "
        "[green]install[/green], [green]list[/green], [green]use[/green], [green]pin[/green], "
        "[green]current[/green], [green]which[/green], [green]exec[/green], "
        "[green]uninstall[/green], [green]releases[/green].\n"
        "For install you can use either [green]opavm install regal 0.38.1[/green] "
        "or [green]opavm install 0.38.1 --tool regal[/green]."
    ),
    rich_markup_mode="rich",
)
console = Console()
OPA_COMMANDS = [
    ("bench", "Benchmark a Rego query"),
    ("build", "Build an OPA bundle"),
    ("capabilities", "Print the capabilities of OPA"),
    ("check", "Check Rego source files"),
    ("completion", "Generate shell completion"),
    ("deps", "Analyze Rego query dependencies"),
    ("eval", "Evaluate a Rego query"),
    ("exec", "Execute against input files"),
    ("fmt", "Format Rego source files"),
    ("help", "Help about any command"),
    ("inspect", "Inspect OPA bundle(s)"),
    ("parse", "Parse Rego source file"),
    ("run", "Start OPA in interactive or server mode"),
    ("sign", "Generate an OPA bundle signature"),
    ("test", "Execute Rego test cases"),
    ("version", "Print the version of OPA"),
]


def _handle_error(err: OpavmError) -> None:
    console.print(f"[red]{err.format()}[/red]")
    raise typer.Exit(code=1)


def _resolve_install_target(subject: str, version: str | None) -> tuple[catalog.ToolSpec, str]:
    if version is None:
        lowered = subject.lower().strip()
        if lowered in catalog.SUPPORTED_TOOLS:
            return catalog.get_tool(lowered), "latest"
        return catalog.get_tool("opa"), subject
    return catalog.get_tool(subject), version


def _resolve_install_target_with_option(
    subject: str,
    version: str | None,
    tool_option: str | None,
) -> tuple[catalog.ToolSpec, str]:
    if tool_option is None:
        return _resolve_install_target(subject, version)

    spec = catalog.get_tool(tool_option)
    if version is not None:
        raise OpavmError(
            "Invalid install arguments.",
            "With --tool, use: opavm install <version> --tool <opa|regal>",
        )

    if subject.lower().strip() == spec.name:
        return spec, "latest"
    return spec, subject


def _install_with_progress(tool: str, version: str) -> str:
    spec = catalog.get_tool(tool)
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        console=console,
        transient=True,
    ) as progress:
        task_id = progress.add_task("Preparing install...", total=100.0, completed=0.0)

        def on_status(stage: str) -> None:
            if stage == "resolving":
                progress.update(
                    task_id, description=f"Resolving {spec.display_name} release metadata...", completed=5.0
                )
            elif stage == "downloading":
                progress.update(
                    task_id, description=f"Downloading {spec.display_name} binary...", completed=10.0
                )
            elif stage == "verifying":
                progress.update(task_id, description=f"Verifying {spec.display_name} binary...", completed=95.0)
            elif stage == "verifying_checksum":
                progress.update(task_id, description=f"Verifying {spec.display_name} checksum...", completed=92.0)
            elif stage == "done":
                progress.update(task_id, description="Install complete.", completed=100.0)
            elif stage == "already_installed":
                progress.update(task_id, description="Version already installed.", completed=100.0)

        def on_download(total_bytes: int | None, downloaded_bytes: int) -> None:
            if total_bytes and total_bytes > 0:
                ratio = min(downloaded_bytes / total_bytes, 1.0)
                progress_pct = 10.0 + (ratio * 80.0)
                progress.update(
                    task_id, description=f"Downloading {spec.display_name} binary...", completed=progress_pct
                )
                return
            # Unknown content length: keep moving the bar while showing bytes received.
            task = progress.tasks[task_id]
            next_progress = task.completed + 1.0
            if next_progress > 90.0:
                next_progress = 10.0
            progress.update(
                task_id,
                description=f"Downloading {spec.display_name} binary... {decimal(downloaded_bytes)}",
                completed=next_progress,
            )

        return installer.install(version, tool=spec.name, on_status=on_status, on_download=on_download)


def _render_exec_help(spec: catalog.ToolSpec) -> None:
    console.print()
    if spec.name == "opa":
        console.print("Usage: opavm exec -- <opa-args>")
    else:
        console.print(f"Usage: opavm exec --tool {spec.name} -- <{spec.name}-args>")
    console.print(
        f"Resolves {spec.display_name} version ({spec.pin_filename} > global default) and forwards args."
    )
    console.print()
    console.print("Examples")
    if spec.name == "opa":
        console.print("opavm exec -- version")
        console.print("opavm exec -- test -v ./policy")
        console.print('opavm exec -- eval -i input.json -d policy.rego "data.example.allow"')
    else:
        console.print("opavm exec --tool regal -- version")
        console.print("opavm exec --tool regal -- lint policy/")
    console.print()

    if spec.name == "opa":
        table = Table(title="Available OPA Commands", box=box.SIMPLE_HEAVY, show_header=True)
        table.add_column("Command", style="cyan")
        table.add_column("Description")
        for command, description in OPA_COMMANDS:
            table.add_row(command, description)
        console.print(table)


@app.command()
def install(
    subject: str = typer.Argument(..., help="Version for OPA, or tool name like `regal`."),
    version: str | None = typer.Argument(None, help="Version when tool is provided."),
    tool: str | None = typer.Option(
        None,
        "--tool",
        "-t",
        help="[cyan]Target tool[/cyan]: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Install OPA or Regal.

    [bold cyan]Examples[/]
    [green]opavm install 1.13.1[/green]
    [green]opavm install latest[/green]
    [green]opavm install regal 0.38.1[/green]
    [green]opavm install 0.38.1 --tool regal[/green]
    """
    try:
        spec, target_version = _resolve_install_target_with_option(subject, version, tool)
        resolved = _install_with_progress(spec.name, target_version)
        if spec.name == "opa":
            shim_path = shim.ensure_shim()
    except OpavmError as err:
        _handle_error(err)
    console.print(f"Installed {spec.display_name} {resolved}.")
    if spec.name == "opa":
        console.print(f"Shim ready at {shim_path}.")
    else:
        console.print(f"Binary path: {installer.binary_path(resolved, tool=spec.name)}")


@app.command("list")
def list_versions(
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to list: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """List installed tool versions."""
    try:
        spec = catalog.get_tool(tool)
    except OpavmError as err:
        _handle_error(err)
    versions = installer.installed_versions(tool=spec.name)
    if not versions:
        if spec.name == "opa":
            console.print("No installed versions. Run: opavm install latest")
        else:
            console.print(f"No installed {spec.display_name} versions. Run: opavm install {spec.name} latest")
        return
    for version in versions:
        console.print(version)


@app.command()
def use(
    version: str,
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to set global default for: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Set global default tool version."""
    try:
        spec = catalog.get_tool(tool)
        if not installer.is_installed(version, tool=spec.name):
            install_hint = (
                f"Run: opavm install {version}"
                if spec.name == "opa"
                else f"Run: opavm install {spec.name} {version}"
            )
            raise VersionNotInstalledError(
                "Version not installed.",
                install_hint,
            )
        config.set_global_default(spec.name, version)
    except OpavmError as err:
        _handle_error(err)
    console.print(f"Global default for {spec.display_name} set to {version}.")


@app.command()
def pin(
    version: str,
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to pin in current project: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Pin tool version for current project."""
    try:
        spec = catalog.get_tool(tool)
        pinned_version = version
        if not installer.is_installed(version, tool=spec.name):
            should_install = typer.confirm(
                f"{spec.display_name} version {version} is not installed. Install now?",
                default=True,
            )
            if not should_install:
                install_hint = (
                    f"Run: opavm install {version}"
                    if spec.name == "opa"
                    else f"Run: opavm install {spec.name} {version}"
                )
                raise VersionNotInstalledError(
                    "Version not installed.",
                    install_hint,
                )
            pinned_version = _install_with_progress(spec.name, version)
        pin_file = Path.cwd() / spec.pin_filename
        pin_file.write_text(f"{pinned_version}\n", encoding="utf-8")
    except OpavmError as err:
        _handle_error(err)
    console.print(f"Pinned {spec.display_name} {pinned_version} in {pin_file}.")


@app.command()
def current(
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to resolve current version for: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Show active tool version and resolution reason."""
    try:
        spec = catalog.get_tool(tool)
        version, reason = resolver.resolve_version(Path.cwd(), tool=spec.name)
    except OpavmError as err:
        _handle_error(err)
    console.print(f"{spec.display_name} {version} ({reason})")


@app.command()
def which(
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to print binary path for: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Print resolved tool binary path."""
    try:
        spec = catalog.get_tool(tool)
        _, _, binary = runner.resolved_binary_path(Path.cwd(), tool=spec.name)
    except OpavmError as err:
        _handle_error(err)
    typer.echo(str(binary.resolve()))


@app.command(
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        "help_option_names": [],
    }
)
def exec(
    ctx: typer.Context,
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to execute: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Run tool through resolved version and forward args."""
    try:
        spec = catalog.get_tool(tool)
    except OpavmError as err:
        _handle_error(err)

    if len(ctx.args) == 1 and ctx.args[0] in {"-h", "--help", "help"}:
        _render_exec_help(spec)
        raise typer.Exit(code=0)

    try:
        _, _, binary = runner.resolved_binary_path(Path.cwd(), tool=spec.name)
    except OpavmError as err:
        _handle_error(err)

    cmd = [str(binary), *ctx.args]
    try:
        proc = subprocess.run(cmd)
    except OSError as exc:
        _handle_error(OpavmError(f"Failed to execute {spec.display_name}.", str(exc)))
    raise typer.Exit(proc.returncode)


@app.command()
def uninstall(
    version: str,
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool to uninstall: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Uninstall a specific tool version."""
    try:
        spec = catalog.get_tool(tool)
        installer.uninstall(version, tool=spec.name)
    except OpavmError as err:
        _handle_error(err)
    console.print(f"Uninstalled {spec.display_name} {version}.")


@app.command("releases")
def releases(
    limit: int = typer.Option(10, "--limit", min=1, help="Number of recent releases"),
    tool: str = typer.Option(
        "opa",
        "--tool",
        "-t",
        help="Tool releases to show: `opa` or `regal`.",
        rich_help_panel="Tool Selection",
    ),
) -> None:
    """Show recent tool releases from GitHub."""
    try:
        spec = catalog.get_tool(tool)
        repo = spec.default_repo
        releases_data = github.fetch_recent_releases(limit=limit, repo=repo)
    except OpavmError as err:
        _handle_error(err)

    console.print()
    console.print(f"{spec.display_name} Releases")
    console.print(f"https://github.com/{repo}/releases")
    console.print()

    table = Table(box=box.SIMPLE_HEAVY, show_header=True, header_style="bold")
    table.add_column("Version", style="cyan")
    table.add_column("Tag")
    table.add_column("Published")
    table.add_column("Pre-release", justify="center")

    for item in releases_data:
        published = item.published_at[:10] if item.published_at else "-"
        table.add_row(item.version, item.tag, published, "yes" if item.prerelease else "no")

    console.print(table)



from __future__ import annotations

from dataclasses import dataclass


class OpavmError(Exception):
    """Base exception with user-facing remediation text."""

    def __init__(self, message: str, hint: str | None = None) -> None:
        super().__init__(message)
        self.message = message
        self.hint = hint

    def format(self) -> str:
        if self.hint:
            return f"{self.message} {self.hint}"
        return self.message


@dataclass
class ResolutionResult:
    version: str
    reason: str


class UnsupportedPlatformError(OpavmError):
    pass


class VersionNotConfiguredError(OpavmError):
    pass


class VersionNotInstalledError(OpavmError):
    pass


class DownloadError(OpavmError):
    pass


class GitHubLookupError(OpavmError):
    pass



from __future__ import annotations

from pathlib import Path

import httpx
from typer.testing import CliRunner

from opavm import cli


class _RateLimitedClient:
    def __init__(self, *_args, **_kwargs) -> None:
        return None

    def __enter__(self) -> "_RateLimitedClient":
        return self

    def __exit__(self, _exc_type, _exc, _tb) -> bool:
        return False

    def get(self, url: str, headers: dict[str, str]) -> httpx.Response:
        _ = headers
        request = httpx.Request("GET", url)
        response = httpx.Response(403, request=request, headers={"x-ratelimit-remaining": "0"})
        raise httpx.HTTPStatusError("rate limited", request=request, response=response)


class _ProxyErrorClient:
    def __init__(self, *_args, **_kwargs) -> None:
        return None

    def __enter__(self) -> "_ProxyErrorClient":
        return self

    def __exit__(self, _exc_type, _exc, _tb) -> bool:
        return False

    def get(self, url: str, headers: dict[str, str]) -> httpx.Response:
        _ = headers
        request = httpx.Request("GET", url)
        raise httpx.ProxyError("proxy failure", request=request)


class _ConnectErrorClient:
    def __init__(self, *_args, **_kwargs) -> None:
        return None

    def __enter__(self) -> "_ConnectErrorClient":
        return self

    def __exit__(self, _exc_type, _exc, _tb) -> bool:
        return False

    def get(self, url: str, headers: dict[str, str]) -> httpx.Response:
        _ = headers
        request = httpx.Request("GET", url)
        raise httpx.ConnectError("connect failure", request=request)


def test_cli_releases_rate_limited_message(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr("opavm.github.httpx.Client", _RateLimitedClient)

    result = runner.invoke(cli.app, ["releases", "--limit", "1"])

    assert result.exit_code == 1
    assert "rate limit" in result.output.lower()


def test_cli_releases_proxy_error_message(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr("opavm.github.httpx.Client", _ProxyErrorClient)

    result = runner.invoke(cli.app, ["releases", "--limit", "1"])

    assert result.exit_code == 1
    assert "proxy" in result.output.lower()


def test_cli_install_network_fault_message(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr("opavm.github.httpx.Client", _ConnectErrorClient)
    monkeypatch.setattr("opavm.installer.platform.normalized_os_arch", lambda: ("linux", "amd64"))

    result = runner.invoke(cli.app, ["install", "latest"])

    assert result.exit_code == 1
    assert "network" in result.output.lower() or "connect" in result.output.lower()



from __future__ import annotations

import pytest

from opavm import github
from opavm.errors import GitHubLookupError


def test_configured_repo_defaults() -> None:
    assert github.configured_repo() == "open-policy-agent/opa"


def test_configured_repo_custom_default() -> None:
    assert github.configured_repo("StyraInc/regal") == "StyraInc/regal"


def test_configured_repo_invalid_default() -> None:
    with pytest.raises(GitHubLookupError):
        github.configured_repo("not-valid")


def test_fetch_release_uses_explicit_repo(monkeypatch: pytest.MonkeyPatch) -> None:
    seen: dict[str, str] = {}

    class FakeResponse:
        def raise_for_status(self) -> None:
            return None

        def json(self) -> dict[str, object]:
            return {"tag_name": "v1.2.3", "assets": []}

    class FakeClient:
        def __init__(self, *_args, **_kwargs) -> None:
            return None

        def __enter__(self) -> "FakeClient":
            return self

        def __exit__(self, _exc_type, _exc, _tb) -> bool:
            return False

        def get(self, url: str, headers: dict[str, str]) -> FakeResponse:
            seen["url"] = url
            seen["accept"] = headers.get("Accept", "")
            return FakeResponse()

    monkeypatch.setattr(github.httpx, "Client", FakeClient)

    release = github.fetch_release("1.2.3", repo="acme/custom-opa")
    assert release.version == "1.2.3"
    assert seen["url"] == "https://api.github.com/repos/acme/custom-opa/releases/tags/v1.2.3"
    assert seen["accept"] == "application/vnd.github+json"


def test_fetch_recent_releases_uses_explicit_repo(monkeypatch: pytest.MonkeyPatch) -> None:
    seen: dict[str, str] = {}

    class FakeResponse:
        def raise_for_status(self) -> None:
            return None

        def json(self) -> list[dict[str, object]]:
            return [
                {
                    "tag_name": "v1.2.3",
                    "published_at": "2026-02-01T00:00:00Z",
                    "prerelease": False,
                }
            ]

    class FakeClient:
        def __init__(self, *_args, **_kwargs) -> None:
            return None

        def __enter__(self) -> "FakeClient":
            return self

        def __exit__(self, _exc_type, _exc, _tb) -> bool:
            return False

        def get(self, url: str, headers: dict[str, str]) -> FakeResponse:
            seen["url"] = url
            seen["accept"] = headers.get("Accept", "")
            return FakeResponse()

    monkeypatch.setattr(github.httpx, "Client", FakeClient)

    releases = github.fetch_recent_releases(limit=1, repo="acme/custom-opa")
    assert releases[0].version == "1.2.3"
    assert seen["url"] == "https://api.github.com/repos/acme/custom-opa/releases?per_page=1"
    assert seen["accept"] == "application/vnd.github+json"


def test_pick_asset_url_prefers_first_candidate() -> None:
    release = github.ReleaseInfo(
        version="1.2.3",
        tag="v1.2.3",
        assets=[
            github.ReleaseAsset(name="opa_linux_arm64_static", url="https://example/static"),
            github.ReleaseAsset(name="opa_linux_amd64", url="https://example/amd64"),
        ],
    )
    picked = github.pick_asset_url(release, ["opa_linux_arm64", "opa_linux_arm64_static"])
    assert picked == "https://example/static"



from __future__ import annotations

from dataclasses import dataclass

from opavm.errors import OpavmError


@dataclass(frozen=True)
class ToolSpec:
    name: str
    display_name: str
    binary_base: str
    default_repo: str
    pin_filename: str


SUPPORTED_TOOLS: dict[str, ToolSpec] = {
    "opa": ToolSpec(
        name="opa",
        display_name="OPA",
        binary_base="opa",
        default_repo="open-policy-agent/opa",
        pin_filename=".opa-version",
    ),
    "regal": ToolSpec(
        name="regal",
        display_name="Regal",
        binary_base="regal",
        default_repo="StyraInc/regal",
        pin_filename=".regal-version",
    ),
}


def get_tool(tool: str) -> ToolSpec:
    normalized = tool.lower().strip()
    spec = SUPPORTED_TOOLS.get(normalized)
    if spec is None:
        options = ", ".join(sorted(SUPPORTED_TOOLS))
        raise OpavmError("Unknown tool.", f"Supported tools: {options}")
    return spec



from __future__ import annotations

from pathlib import Path

from opavm import catalog, config
from opavm.errors import VersionNotConfiguredError


def pin_filename(tool: str = "opa") -> str:
    spec = catalog.get_tool(tool)
    return spec.pin_filename


def find_pin_file(start: Path, tool: str = "opa") -> Path | None:
    current = start.resolve()
    filename = pin_filename(tool)
    for directory in [current, *current.parents]:
        pin = directory / filename
        if pin.exists():
            return pin
    return None


def resolve_version(start: Path, tool: str = "opa") -> tuple[str, str]:
    spec = catalog.get_tool(tool)
    pin_file = find_pin_file(start, tool=spec.name)
    if pin_file:
        pinned = pin_file.read_text(encoding="utf-8").strip()
        if pinned:
            return pinned, f"pinned via {pin_file}"

    global_default = config.get_global_default(spec.name)
    if global_default:
        return global_default, "global default"

    use_hint = "Run: opavm use <version>"
    pin_hint = f"create {spec.pin_filename}"
    if spec.name != "opa":
        use_hint = f"Run: opavm use <version> --tool {spec.name}"

    raise VersionNotConfiguredError(
        "No version configured.",
        f"{use_hint} or {pin_hint}.",
    )



from __future__ import annotations

from pathlib import Path

import pytest

from opavm import config, shim


def test_ensure_shim_windows_creates_cmd(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    monkeypatch.setattr("opavm.shim.platform.normalized_os_arch", lambda: ("windows", "amd64"))

    shim_path = shim.ensure_shim()

    assert shim_path.name == "opa.cmd"
    content = shim_path.read_text(encoding="utf-8")
    assert "opavm which" in content
    assert "%*" in content


def test_ensure_shim_posix_creates_opa(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    monkeypatch.setattr("opavm.shim.platform.normalized_os_arch", lambda: ("linux", "amd64"))

    shim_path = shim.ensure_shim()

    assert shim_path.name == "opa"
    assert (config.shims_dir() / "opa").exists()
    content = shim_path.read_text(encoding="utf-8")
    assert "opavm which" in content



from __future__ import annotations

import platform

from opavm.errors import UnsupportedPlatformError


def normalized_os_arch() -> tuple[str, str]:
    sys_name = platform.system().lower()
    machine = platform.machine().lower()

    if sys_name == "darwin":
        os_name = "darwin"
    elif sys_name == "linux":
        os_name = "linux"
    elif sys_name == "windows":
        os_name = "windows"
    else:
        raise UnsupportedPlatformError(
            f"Unsupported OS: {sys_name}.", "opavm supports macOS, Linux, and Windows."
        )

    if machine in {"x86_64", "amd64"}:
        arch = "amd64"
    elif machine in {"aarch64", "arm64"}:
        arch = "arm64"
    else:
        raise UnsupportedPlatformError(
            f"Unsupported architecture: {machine}.", "Use amd64 or arm64 hardware."
        )

    if os_name == "windows" and arch != "amd64":
        raise UnsupportedPlatformError(
            f"Unsupported architecture: {machine}.",
            "Windows support currently requires amd64.",
        )

    return os_name, arch


def asset_name(version: str, os_name: str, arch: str, binary_base: str = "opa") -> str:
    _ = version
    extension = ".exe" if os_name == "windows" else ""
    return f"{binary_base}_{os_name}_{arch}{extension}"


def asset_name_candidates(os_name: str, arch: str, binary_base: str = "opa") -> list[str]:
    primary = asset_name("latest", os_name, arch, binary_base=binary_base)
    candidates = [primary]

    # Older arm64 OPA releases expose only static assets on macOS/Linux.
    if binary_base == "opa" and os_name in {"darwin", "linux"} and arch == "arm64":
        candidates.append(f"opa_{os_name}_{arch}_static")

    return candidates


def binary_filename(os_name: str, binary_base: str = "opa") -> str:
    return f"{binary_base}.exe" if os_name == "windows" else binary_base



from __future__ import annotations

from pathlib import Path

from opavm import config, platform


def ensure_shim() -> Path:
    config.ensure_layout()
    os_name, _ = platform.normalized_os_arch()

    if os_name == "windows":
        shim_path = config.shims_dir() / "opa.cmd"
        if shim_path.exists():
            return shim_path
        content = """@echo off
for /f "delims=" %%i in ('opavm which') do set "OPA_BIN=%%i"
"%OPA_BIN%" %*
"""
        shim_path.write_text(content, encoding="utf-8")
        return shim_path

    shim_path = config.shims_dir() / "opa"
    if shim_path.exists():
        return shim_path
    content = """#!/usr/bin/env sh
set -eu
resolved="$(opavm which)"
exec "$resolved" "$@"
"""
    shim_path.write_text(content, encoding="utf-8")
    shim_path.chmod(shim_path.stat().st_mode | 0o111)
    return shim_path


def path_instruction() -> str:
    os_name, _ = platform.normalized_os_arch()
    if os_name == "windows":
        return f'$env:Path = "{config.shims_dir()};$env:Path"'
    return f'export PATH="{config.shims_dir()}:$PATH"'



from __future__ import annotations

from dataclasses import dataclass
import os
from typing import Sequence

import httpx

from opavm.errors import GitHubLookupError

GITHUB_API_ROOT = "https://api.github.com/repos"


@dataclass
class ReleaseAsset:
    name: str
    url: str


@dataclass
class ReleaseInfo:
    version: str
    tag: str
    assets: list[ReleaseAsset]


@dataclass
class ReleaseSummary:
    version: str
    tag: str
    published_at: str
    prerelease: bool


def _github_headers() -> dict[str, str]:
    headers = {"Accept": "application/vnd.github+json"}
    token = os.environ.get("OPAVM_GITHUB_TOKEN", "").strip()
    if token:
        headers["Authorization"] = f"Bearer {token}"
    return headers


def _raise_friendly_http_error(exc: httpx.HTTPError) -> GitHubLookupError:
    if isinstance(exc, httpx.ProxyError):
        return GitHubLookupError(
            "Failed to query GitHub releases.",
            "Proxy error. Check HTTP_PROXY/HTTPS_PROXY or corporate proxy settings.",
        )
    if isinstance(exc, httpx.ConnectError):
        return GitHubLookupError(
            "Failed to query GitHub releases.",
            "Network connection failed. Check internet connectivity, DNS, firewall, or VPN.",
        )
    if isinstance(exc, httpx.HTTPStatusError):
        status = exc.response.status_code
        if status == 403 and exc.response.headers.get("x-ratelimit-remaining") == "0":
            return GitHubLookupError(
                "GitHub API rate limit exceeded.",
                "Set OPAVM_GITHUB_TOKEN or retry after the rate limit resets.",
            )
        if status == 404:
            return GitHubLookupError(
                "Requested release was not found.",
                "Check version tag and access permissions.",
            )
        if status in {401, 403}:
            return GitHubLookupError(
                "GitHub request was unauthorized.",
                "Check OPAVM_GITHUB_TOKEN and repository access.",
            )
        return GitHubLookupError(
            "Failed to query GitHub releases.",
            f"GitHub API responded with HTTP {status}.",
        )
    return GitHubLookupError(
        "Failed to query GitHub releases.",
        "Check network connectivity, proxy settings, or GitHub availability.",
    )


def configured_repo(
    default_repo: str = "open-policy-agent/opa",
) -> str:
    repo = default_repo.strip()
    parts = repo.split("/")
    if len(parts) != 2 or not parts[0] or not parts[1]:
        raise GitHubLookupError(
            "Invalid repository value.",
            f"Use format: owner/repo (example: {default_repo}).",
        )
    return repo


def validate_repo(repo: str) -> str:
    normalized = repo.strip()
    parts = normalized.split("/")
    if len(parts) != 2 or not parts[0] or not parts[1]:
        raise GitHubLookupError(
            "Invalid repository value.",
            "Use format: owner/repo (example: open-policy-agent/opa).",
        )
    return normalized


def releases_api_url(repo: str) -> str:
    return f"{GITHUB_API_ROOT}/{repo}/releases"


def _normalize_tag(version: str) -> str:
    if version == "latest":
        return version
    if version.startswith("v"):
        return version
    return f"v{version}"


def _version_from_tag(tag: str) -> str:
    return tag[1:] if tag.startswith("v") else tag


def fetch_release(version: str, timeout: float = 30.0, repo: str | None = None) -> ReleaseInfo:
    tag = _normalize_tag(version)
    selected_repo = validate_repo(repo) if repo else configured_repo()
    base_url = releases_api_url(selected_repo)
    url = f"{base_url}/latest" if version == "latest" else f"{base_url}/tags/{tag}"

    try:
        with httpx.Client(timeout=timeout, follow_redirects=True) as client:
            response = client.get(url, headers=_github_headers())
            response.raise_for_status()
            payload = response.json()
    except httpx.HTTPError as exc:
        raise _raise_friendly_http_error(exc) from exc

    release_tag = payload.get("tag_name")
    if not release_tag:
        raise GitHubLookupError("Invalid GitHub response.", "Release tag missing.")

    assets = [
        ReleaseAsset(name=item["name"], url=item["browser_download_url"])
        for item in payload.get("assets", [])
        if "name" in item and "browser_download_url" in item
    ]
    return ReleaseInfo(version=_version_from_tag(release_tag), tag=release_tag, assets=assets)


def fetch_recent_releases(
    limit: int = 10,
    timeout: float = 30.0,
    repo: str | None = None,
) -> list[ReleaseSummary]:
    if limit < 1:
        raise GitHubLookupError("Limit must be at least 1.", "Try: opavm releases --limit 10")

    selected_repo = validate_repo(repo) if repo else configured_repo()
    base_url = releases_api_url(selected_repo)
    url = f"{base_url}?per_page={min(limit, 100)}"
    try:
        with httpx.Client(timeout=timeout, follow_redirects=True) as client:
            response = client.get(url, headers=_github_headers())
            response.raise_for_status()
            payload = response.json()
    except httpx.HTTPError as exc:
        raise _raise_friendly_http_error(exc) from exc

    if not isinstance(payload, list):
        raise GitHubLookupError("Invalid GitHub response.", "Expected release list.")

    releases: list[ReleaseSummary] = []
    for item in payload:
        tag = item.get("tag_name")
        if not tag:
            continue
        releases.append(
            ReleaseSummary(
                version=_version_from_tag(tag),
                tag=tag,
                published_at=str(item.get("published_at") or ""),
                prerelease=bool(item.get("prerelease")),
            )
        )
        if len(releases) >= limit:
            break

    return releases


def pick_asset_url(release: ReleaseInfo, expected_names: Sequence[str]) -> str:
    for expected_name in expected_names:
        for asset in release.assets:
            if asset.name == expected_name:
                return asset.url
    expected = ", ".join(expected_names)
    raise GitHubLookupError(f"No matching asset found for {expected}.", "Check available release assets.")


def checksum_asset_url(release: ReleaseInfo, asset_name: str) -> str | None:
    checksum_name = f"{asset_name}.sha256"
    raw_assets = getattr(release, "assets", [])
    try:
        assets = list(raw_assets)
    except TypeError:
        return None
    for asset in assets:
        if asset.name == checksum_name:
            return asset.url
    return None



from __future__ import annotations

from pathlib import Path
from unittest import mock

import pytest

from opavm import config, installer


def test_install_layout_and_idempotent(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    fake_release = mock.Mock(version="0.62.1")

    with mock.patch("opavm.installer.platform.normalized_os_arch", return_value=("linux", "amd64")), mock.patch(
        "opavm.installer.github.fetch_release", return_value=fake_release
    ), mock.patch("opavm.installer.github.pick_asset_url", return_value="https://example.test/opa"), mock.patch(
        "opavm.installer.download.download_binary"
    ) as mock_download, mock.patch("opavm.installer.verify_binary") as mock_verify:
        def fake_download(_url: str, destination: Path, on_progress=None) -> None:
            destination.parent.mkdir(parents=True, exist_ok=True)
            destination.write_text("bin", encoding="utf-8")
            if on_progress is not None:
                on_progress(3, 3)

        mock_download.side_effect = fake_download

        installed = installer.install("0.62.1")
        assert installed == "0.62.1"
        assert (config.versions_dir() / "0.62.1" / "opa").exists()
        assert mock_verify.call_count == 1

        installed_again = installer.install("0.62.1")
        assert installed_again == "0.62.1"
        assert mock_download.call_count == 1


def test_install_regal_layout(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    fake_release = mock.Mock(version="0.38.1")

    with mock.patch("opavm.installer.platform.normalized_os_arch", return_value=("linux", "amd64")), mock.patch(
        "opavm.installer.github.fetch_release", return_value=fake_release
    ) as fetch_release_mock, mock.patch(
        "opavm.installer.github.pick_asset_url", return_value="https://example.test/regal"
    ) as pick_mock, mock.patch("opavm.installer.download.download_binary") as mock_download, mock.patch(
        "opavm.installer.verify_binary"
    ):
        def fake_download(_url: str, destination: Path, on_progress=None) -> None:
            destination.parent.mkdir(parents=True, exist_ok=True)
            destination.write_text("bin", encoding="utf-8")
            if on_progress is not None:
                on_progress(3, 3)

        mock_download.side_effect = fake_download

        installed = installer.install("0.38.1", tool="regal")
        assert installed == "0.38.1"
        assert (config.base_dir() / "tools" / "regal" / "versions" / "0.38.1" / "regal").exists()
        expected_candidates = ["regal_Linux_x86_64"]
        assert pick_mock.call_args.args[1] == expected_candidates
        assert fetch_release_mock.call_args.kwargs["repo"] == "StyraInc/regal"



from __future__ import annotations

from pathlib import Path
from unittest import mock

from typer.testing import CliRunner

from opavm import cli, github, installer


def test_cli_root_help_includes_tool_selection(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    result = runner.invoke(cli.app, ["--help"])

    assert result.exit_code == 0
    assert "Tool Selection" in result.output
    assert "--tool" in result.output
    assert "opa" in result.output
    assert "regal" in result.output


def test_cli_current_smoke(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    (tmp_path / ".opa-version").write_text("0.62.1\n", encoding="utf-8")

    result = runner.invoke(cli.app, ["current"])
    assert result.exit_code == 0
    assert "OPA 0.62.1" in result.output


def test_cli_which_smoke(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    opavm_home = tmp_path / ".opavm"
    monkeypatch.setenv("OPAVM_HOME", str(opavm_home))

    (tmp_path / ".opa-version").write_text("0.62.1\n", encoding="utf-8")
    binary = installer.binary_path("0.62.1")
    binary.parent.mkdir(parents=True)
    binary.write_text("fake", encoding="utf-8")

    result = runner.invoke(cli.app, ["which"])
    assert result.exit_code == 0
    assert str(binary.resolve()) in result.output


def test_cli_exec_forwards_args_and_exit_code(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    opavm_home = tmp_path / ".opavm"
    monkeypatch.setenv("OPAVM_HOME", str(opavm_home))

    (tmp_path / ".opa-version").write_text("0.62.1\n", encoding="utf-8")
    binary = installer.binary_path("0.62.1")
    binary.parent.mkdir(parents=True)
    binary.write_text("fake", encoding="utf-8")

    with mock.patch("opavm.cli.subprocess.run", return_value=mock.Mock(returncode=7)) as run_mock:
        result = runner.invoke(cli.app, ["exec", "--", "version"])

    assert result.exit_code == 7
    run_mock.assert_called_once_with([str(binary), "version"])


def test_cli_pin_prompts_install_for_missing_version(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    with mock.patch("opavm.cli.installer.is_installed", return_value=False), mock.patch(
        "opavm.cli.installer.install", return_value="0.61.0"
    ) as install_mock:
        result = runner.invoke(cli.app, ["pin", "0.61.0"], input="y\n")

    assert result.exit_code == 0
    assert install_mock.call_count == 1
    args, kwargs = install_mock.call_args
    assert args == ("0.61.0",)
    assert kwargs["tool"] == "opa"
    assert callable(kwargs["on_status"])
    assert callable(kwargs["on_download"])
    assert (tmp_path / ".opa-version").read_text(encoding="utf-8") == "0.61.0\n"


def test_cli_pin_decline_install_exits_with_error(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    with mock.patch("opavm.cli.installer.is_installed", return_value=False), mock.patch(
        "opavm.cli.installer.install"
    ) as install_mock:
        result = runner.invoke(cli.app, ["pin", "0.61.0"], input="n\n")

    assert result.exit_code == 1
    install_mock.assert_not_called()
    assert not (tmp_path / ".opa-version").exists()


def test_cli_releases_table(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    releases_data = [
        github.ReleaseSummary(
            version="1.2.3",
            tag="v1.2.3",
            published_at="2026-02-01T00:00:00Z",
            prerelease=False,
        )
    ]
    with mock.patch("opavm.cli.github.fetch_recent_releases", return_value=releases_data):
        result = runner.invoke(cli.app, ["releases", "--limit", "1"])

    assert result.exit_code == 0
    assert result.output.startswith("\n")
    assert "OPA Releases" in result.output
    assert "https://github.com/open-policy-agent/opa/releases" in result.output
    assert "1.2.3" in result.output


def test_cli_releases_regal_table_header_link(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    releases_data = [
        github.ReleaseSummary(
            version="0.38.1",
            tag="v0.38.1",
            published_at="2026-02-01T00:00:00Z",
            prerelease=False,
        )
    ]
    with mock.patch("opavm.cli.github.fetch_recent_releases", return_value=releases_data):
        result = runner.invoke(cli.app, ["releases", "--tool", "regal", "--limit", "1"])

    assert result.exit_code == 0
    assert "Regal Releases" in result.output
    assert "https://github.com/StyraInc/regal/releases" in result.output
    assert "0.38.1" in result.output


def test_cli_install_help_includes_tool_and_examples(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    result = runner.invoke(cli.app, ["install", "--help"])

    assert result.exit_code == 0
    assert "--tool" in result.output
    assert "Tool Selection" in result.output
    assert "Examples" in result.output
    assert "opavm install regal 0.38.1" in result.output


def test_cli_exec_help_shows_opa_commands(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    result = runner.invoke(cli.app, ["exec", "--help"])

    assert result.exit_code == 0
    assert "Usage: opavm exec -- <opa-args>" in result.output
    assert "Available OPA Commands" in result.output
    assert "bench" in result.output
    assert "test" in result.output


def test_cli_use_regal_sets_global_default(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    with mock.patch("opavm.cli.installer.is_installed", return_value=True):
        result = runner.invoke(cli.app, ["use", "0.38.1", "--tool", "regal"])

    assert result.exit_code == 0
    assert "Global default for Regal set to 0.38.1." in result.output


def test_cli_pin_regal_writes_regal_pin_file(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    with mock.patch("opavm.cli.installer.is_installed", return_value=True):
        result = runner.invoke(cli.app, ["pin", "0.38.1", "--tool", "regal"])

    assert result.exit_code == 0
    assert (tmp_path / ".regal-version").read_text(encoding="utf-8") == "0.38.1\n"


def test_cli_current_regal_smoke(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    (tmp_path / ".regal-version").write_text("0.38.1\n", encoding="utf-8")

    result = runner.invoke(cli.app, ["current", "--tool", "regal"])
    assert result.exit_code == 0
    assert "Regal 0.38.1" in result.output


def test_cli_which_regal_smoke(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    opavm_home = tmp_path / ".opavm"
    monkeypatch.setenv("OPAVM_HOME", str(opavm_home))

    (tmp_path / ".regal-version").write_text("0.38.1\n", encoding="utf-8")
    binary = installer.binary_path("0.38.1", tool="regal")
    binary.parent.mkdir(parents=True)
    binary.write_text("fake", encoding="utf-8")

    result = runner.invoke(cli.app, ["which", "--tool", "regal"])
    assert result.exit_code == 0
    assert str(binary.resolve()) in result.output


def test_cli_exec_regal_forwards_args_and_exit_code(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    opavm_home = tmp_path / ".opavm"
    monkeypatch.setenv("OPAVM_HOME", str(opavm_home))

    (tmp_path / ".regal-version").write_text("0.38.1\n", encoding="utf-8")
    binary = installer.binary_path("0.38.1", tool="regal")
    binary.parent.mkdir(parents=True)
    binary.write_text("fake", encoding="utf-8")

    with mock.patch("opavm.cli.subprocess.run", return_value=mock.Mock(returncode=3)) as run_mock:
        result = runner.invoke(cli.app, ["exec", "--tool", "regal", "--", "version"])

    assert result.exit_code == 3
    run_mock.assert_called_once_with([str(binary), "version"])


def test_cli_install_regal_invokes_regal_tool(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    with mock.patch("opavm.cli._install_with_progress", return_value="0.38.1") as install_mock, mock.patch(
        "opavm.cli.installer.binary_path", return_value=Path("/tmp/regal")
    ) as path_mock, mock.patch("opavm.cli.shim.ensure_shim") as shim_mock:
        result = runner.invoke(cli.app, ["install", "regal", "0.38.1"])

    assert result.exit_code == 0
    install_mock.assert_called_once_with("regal", "0.38.1")
    path_mock.assert_called_once_with("0.38.1", tool="regal")
    shim_mock.assert_not_called()
    assert "Installed Regal 0.38.1." in result.output


def test_cli_install_tool_without_version_defaults_latest(tmp_path: Path, monkeypatch) -> None:
    runner = CliRunner()
    monkeypatch.chdir(tmp_path)

    with mock.patch("opavm.cli._install_with_progress", return_value="0.38.1") as install_mock, mock.patch(
        "opavm.cli.installer.binary_path", return_value=Path("/tmp/regal")
    ), mock.patch("opavm.cli.shim.ensure_shim") as shim_mock:
        result = runner.invoke(cli.app, ["install", "regal"])

    assert result.exit_code == 0
    install_mock.assert_called_once_with("regal", "latest")
    shim_mock.assert_not_called()



from __future__ import annotations

from pathlib import Path

import pytest

from opavm import config
from opavm.errors import VersionNotConfiguredError
from opavm.resolver import find_pin_file, resolve_version


def test_find_pin_file_walks_parents(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)
    project = tmp_path / "a" / "b" / "c"
    project.mkdir(parents=True)
    pin = tmp_path / "a" / ".opa-version"
    pin.write_text("0.62.1\n", encoding="utf-8")

    found = find_pin_file(project)
    assert found == pin


def test_find_pin_file_walks_parents_for_regal(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)
    project = tmp_path / "a" / "b" / "c"
    project.mkdir(parents=True)
    pin = tmp_path / "a" / ".regal-version"
    pin.write_text("0.38.1\n", encoding="utf-8")

    found = find_pin_file(project, tool="regal")
    assert found == pin


def test_resolve_prefers_pin(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    config.set_global_default("opa", "0.60.0")

    pin = tmp_path / ".opa-version"
    pin.write_text("0.62.1\n", encoding="utf-8")

    version, reason = resolve_version(tmp_path)
    assert version == "0.62.1"
    assert ".opa-version" in reason


def test_resolve_regal_prefers_pin(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    config.set_global_default("regal", "0.37.0")

    pin = tmp_path / ".regal-version"
    pin.write_text("0.38.1\n", encoding="utf-8")

    version, reason = resolve_version(tmp_path, tool="regal")
    assert version == "0.38.1"
    assert ".regal-version" in reason


def test_resolve_falls_back_to_global(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    config.set_global_default("opa", "0.61.0")

    version, reason = resolve_version(tmp_path)
    assert version == "0.61.0"
    assert reason == "global default"


def test_resolve_regal_falls_back_to_global(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))
    config.set_global_default("regal", "0.38.1")

    version, reason = resolve_version(tmp_path, tool="regal")
    assert version == "0.38.1"
    assert reason == "global default"


def test_resolve_errors_when_unconfigured(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("OPAVM_HOME", str(tmp_path / ".opavm"))

    with pytest.raises(VersionNotConfiguredError):
        resolve_version(tmp_path)



from __future__ import annotations

from pathlib import Path

from opavm import catalog, installer, resolver
from opavm.errors import VersionNotInstalledError


def resolved_binary_path(start: Path, tool: str = "opa") -> tuple[str, str, Path]:
    spec = catalog.get_tool(tool)
    version, reason = resolver.resolve_version(start, tool=spec.name)
    binary = installer.binary_path(version, tool=spec.name)
    if not binary.exists():
        install_hint = (
            f"Run: opavm install {version}"
            if spec.name == "opa"
            else f"Run: opavm install {spec.name} {version}"
        )
        raise VersionNotInstalledError(
            "Version not installed.",
            install_hint,
        )
    return version, reason, binary
